//RotatorMkII.ino - Mini Satellite-Antenna Rotator MkII.
//Copyright (c) 2016,2017 Julie VK3FOWL and Joe VK3YSP
//Released under the GNU General Public License.
//For more information please visit http://www.sarcnet.org
//Release history:
//Release 1: Original release
//Release 2: Added support for hamlib 3.0.1. Added debug mode.
//Release 3:
//  Added a low pass filter to the sensor data to improve calibration.
//  Reinitialised the I2C bus and sensor prior to each read to avoid I2C lockups caused my power glitches
//  Changed Serial to Serial1 for RS-422 operation - Change it back to use this code with the MkI version.
 
#include <Arduino.h>
#include <Wire.h>
#include <EEPROM.h>
#include "Lsm303d.h"
Lsm303d lsm303d;
//Defines
#define WINDUP_LIMIT 450
//#define DEBUG

//Types

struct vector {
  //Vector type comprising three orthogonal components.
  float i, j, k;
};

enum Modes {tracking, monitoring, demonstrating, calibrating, debugging};    //Rotator controller modes

//Constants

const float rad2deg = 57.295779513;   //Degrees per radian

//Motor drive pins
const int azBrkPin = 4;  //Azimuth break
const int azDirPin = 5;  //Azimuth direction
const int azPwmPin = 6;  //Azimuth pwm
const int elBrkPin = 7;  //Elevation break
const int elDirPin = 8;  //Elevation direction
const int elPwmPin = 9;  //Elevation pwm

//Motor drive gains. These set the amount of motor drive close to the set point
const int azGain = 25;   //Azimuth motor gain
const int elGain = 25;   //Elevation motor gain

//Global variables
float az;               //Antenna azimuth
float el;               //Antenna elevation
String line;            //Command line
float azSet;            //Antenna azimuth set point
float elSet;            //Antenna elevation set point
float azLast;           //Last antenna azimuth reading
float elLast;           //Last antenna element reading
float azWindup;         //Antenna windup angle from startup azimuth position
float azOffset;         //Antenna azimuth offset for whole revolutions
bool windup;            //Antenna windup condition
float azSpeed;          //Antenna azimuth motor speed
float elSpeed;          //Antenna elevation motor speed
float azInc;            //AZ increment for demo mode
float elInc;            //EL increment for demo mode
int t;                  //Variable delay period - depends on mode
Modes mode;             //Rotator mode
vector mmax, mmin;      //Magnetometer limit vectors
vector gmax, gmin;      //Accelerometer limit vectors
float D;                //The magnetic declination
vector me, ge, ms, gs;  //Calibration vectors - Keep together for save and restore functions

//Functions

//Vector functions

struct vector cross(vector a, vector b) {
  //Returns the vector cross product of two vectors.
  return {a.j * b.k - a.k * b.j, a.k * b.i - a.i * b.k, a.i * b.j - a.j * b.i};
}

float dot(vector a, vector b) {
  //Returns the scalar dot product of two vectors.
  return a.i * b.i + a.j * b.j + a.k * b.k;
}

float mod(vector a) {
  //Returns the scalar modulus of a vector.
  return sqrt(a.i * a.i + a.j * a.j + a.k * a.k);
}

vector unit(vector a) {
  //Returns the unit vector of a vector.
  float modulus = mod(a);
  return {a.i / modulus, a.j / modulus, a.k / modulus};
}

vector neg(vector a) {
  //Returns the vector negative of a vector.
  return { -a.i, -a.j, -a.k};
}

//Rotator functions

void reset(bool getConfig) {
  //Reset the rotator, initialize its variables and optionally get the stored configuration
  azSet = 0.0;
  elSet = 0.0;
  line = "";
  azLast = 0.0;
  elLast = 0.0;
  azWindup = 0.0;
  azOffset = 0.0;
  azSpeed = 0.0;
  elSpeed = 0.0;
  mode = tracking;
  windup = false;
  if (getConfig) restore(52, (byte*) &gs);
  mmin = { -1, -1, -1};
  mmax = {1, 1, 1};
  gmin = { -1, -1, -1};
  gmax = {1, 1, 1};
  azInc = 0.05;
  elInc = 0.05;
  t = 100;
}

float diffAngle(float a, float b) {
  //Calculate the acute angle between two angles in -180..180 degree format
  float diff = a - b;
  if (diff < -180) diff += 360;
  if (diff > 180) diff -= 360;
  return diff;
}

float lpf(float value, float *last, float alpha) {
  //Low pass filter - Decrease alpha to increase damping factor
  float result = (alpha * value) + *last * (1 - alpha);
  *last = result;
  return result;
}

void drive(float err, int gain, float *spdLast, int brkPin, int dirPin, int pwmPin) {
  //Drive the motors to reduce the antenna pointing errors to zero

  //Ingrease the gain to help reduce small errors to zero
  //if (abs(err) < 1) gain *= 2;

  //Calculate the motor speed: A linear ramp close to zero; constant full speed beyond that.
  float spd = constrain(err * gain, -255, 255);

  //Low pass filter the speed to reduce abrubt changes in speed
  spd = lpf(spd, spdLast, 0.4);

  //Set the motor direction
  digitalWrite(dirPin, (spd > 0) ? HIGH : LOW);
  //Engage the motor break when the error is small to reduce dither, noise and current consumption
  digitalWrite(brkPin, (abs(err) < 0.5) ? HIGH : LOW);
  //Set the motor speed
  analogWrite(pwmPin, (byte)(abs(spd)));
}

void save(int bytes, byte *p) {
  //Save the configuration data to flash memory
  for (int i = 0; i < bytes; i++) EEPROM.write(i, *p++);
  printCal();
}

void restore(int bytes, byte *p) {
  //Restore the configuration data from flash memory
  for (int i = 0; i < bytes; i++) *p++ = EEPROM.read(i);
  printCal();
}

void printDebug(void) {
  //Print raw sensor data
  Serial1.print(lsm303d.mx); Serial1.print(",");
  Serial1.print(lsm303d.my); Serial1.print(",");
  Serial1.print(lsm303d.mz); Serial1.print(",");
  Serial1.print(lsm303d.gx); Serial1.print(",");
  Serial1.print(lsm303d.gy); Serial1.print(",");
  Serial1.println(lsm303d.gz);
}

void printCal(void) {
  //Print the calibration data
  if (mode == calibrating) {
    Serial1.print(D, 1); Serial1.print(",");
    Serial1.print(me.i, 1); Serial1.print(",");
    Serial1.print(me.j, 1); Serial1.print(",");
    Serial1.print(me.k, 1); Serial1.print(",");
    Serial1.print(ge.i, 1); Serial1.print(",");
    Serial1.print(ge.j, 1); Serial1.print(",");
    Serial1.print(ge.k, 1); Serial1.print(",");
    Serial1.print(ms.i, 6); Serial1.print(",");
    Serial1.print(ms.j, 6); Serial1.print(",");
    Serial1.print(ms.k, 6); Serial1.print(",");
    Serial1.print(gs.i, 6); Serial1.print(",");
    Serial1.print(gs.j, 6); Serial1.print(",");
    Serial1.println(gs.k, 6);
  }
}

void printMon(float az, float el, float azSet, float elSet, float azWindup, float azError, float elError) {
  //Print the monitor data
  Serial1.print(az, 0); Serial1.print(",");
  Serial1.print(el, 0); Serial1.print(",");
  Serial1.print(azSet, 0); Serial1.print(",");
  Serial1.print(elSet, 0); Serial1.print(",");
  Serial1.print(azWindup, 0); Serial1.print(",");
  Serial1.print(windup); Serial1.print(",");
  Serial1.print(azError, 0); Serial1.print(",");
  Serial1.println(elError, 0);
}

void printAzEl() {
  //Print the rotator feedback data
  Serial1.print("AZ");
  if (az < 0) Serial1.print((az + 360), 1);
  else Serial1.print(az, 1);
  Serial1.print(" EL");
  Serial1.print(el, 1);
  Serial1.print("\n");
}

void minmax(float a, float *b, bool *c) {
  //Update exisiting maximums or minimums based on current values
  if ((*b > 0) && (a > *b) || (*b < 0) && (a < *b)) {
    *b = a;
    *c = true;
  }
}

void calibrate() {
  //Caculate the 3D scaling factors and errors for the magnetometer and accelerometer
  bool changed = false;
  //Update exisiting maximums and minimums based on current values
  minmax(lsm303d.mx, &mmax.i, &changed);
  minmax(lsm303d.my, &mmax.j, &changed);
  minmax(lsm303d.mz, &mmax.k, &changed);
  minmax(lsm303d.gx, &gmax.i, &changed);
  minmax(lsm303d.gy, &gmax.j, &changed);
  minmax(lsm303d.gz, &gmax.k, &changed);
  minmax(lsm303d.mx, &mmin.i, &changed);
  minmax(lsm303d.my, &mmin.j, &changed);
  minmax(lsm303d.mz, &mmin.k, &changed);
  minmax(lsm303d.gx, &gmin.i, &changed);
  minmax(lsm303d.gy, &gmin.j, &changed);
  minmax(lsm303d.gz, &gmin.k, &changed);
  //Calculate the error vectors
  me = {(mmax.i + mmin.i) / 2, (mmax.j + mmin.j) / 2, (mmax.k + mmin.k) / 2};
  ge = {(gmax.i + gmin.i) / 2, (gmax.j + gmin.j) / 2, (gmax.k + gmin.k) / 2};
  //Caclulate the scaling vectors
  ms = {(mmax.i - mmin.i) / 2, (mmax.j - mmin.j) / 2, (mmax.k - mmin.k) / 2};
  gs = {(gmax.i - gmin.i) / 2, (gmax.j - gmin.j) / 2, (gmax.k - gmin.k) / 2};
  //  ms = unit({(mmax.i - mmin.i), (mmax.j - mmin.j), (mmax.k - mmin.k)});
  //  gs = unit({(gmax.i - gmin.i), (gmax.j - gmin.j), (gmax.k - gmin.k)});
  //Print any changes to facilitate manual calibration
  if (changed) printCal();
}

void getAzEl(float *az, float *el) {
  //Get the antenna azimuth and elevation angles
  //Get the unit vectors for the earth's magnetic and gravitational fields
  //For each component subtract the error and multiply by the scaling factor
  //For each component subtract the error and divide by the scaling factor
  vector M = unit({(lsm303d.mx - me.i) / ms.i, (lsm303d.my - me.j) / ms.j, (lsm303d.mz - me.k) / ms.k});
  vector G = unit({(lsm303d.gx - ge.i) / gs.i, (lsm303d.gy - ge.j) / gs.j, (lsm303d.gz - ge.k) / gs.k});
  //  vector M = unit({(lsm303d.mx - me.i) * ms.i, (lsm303d.my - me.j) * ms.j, (lsm303d.mz - me.k) * ms.k});
  //  vector G = unit({(lsm303d.gx - ge.i) * gs.i, (lsm303d.gy - ge.j) * gs.j, (lsm303d.gz - ge.k) * gs.k});
  const vector X = {1.0, 0.0, 0.0};     //The antenna X vector
  const vector Y = {0.0, 1.0, 0.0};     //The antenna Y (boresight) vector
  const vector Z = {0.0, 0.0, 1.0};     //The antenna Z vector
  //G = neg(G);
  //Compute the antenna pointing angles
  vector E = cross(G, M);               //The magnetic East vector
  vector N = cross(E, G);               //The magnetic North vector
  vector U = neg(G);                    //The magnetic Up vector
  float Xn = dot(X, N);                 //The scalar projection of X onto N
  float Xe = dot(X, E);                 //The scalar projection of X onto E
  float Yu = dot(Y, U);                 //The scalar projection of Y onto U
  float Zu = dot(Z, U);                 //The scalar projection of Z onto U
  *az = atan2(-Xn, Xe) * rad2deg + D;   //The azimuth angle in degrees using the X-axis
  *el = atan2(Yu, Zu) * rad2deg;        //The elevation angle in degrees using the Y-axis
  if (*az > 180) *az = *az - 360;       //Ensure azimuth -180..180 format after adding D
}

void getWindup(bool *windup,  float *azWindup, float *azOffset, float *azLast, float *elLast, float az, float elSet) {
  //Get the accumulated windup angle from the home position. Set the windup state.
  //Get the raw difference angle between the current and last azimuth reading from the sensor
  float azDiff = az - *azLast;

  //Detect crossing South: azDiff jumps 360 for a clockwise crossing or -360 for an anticlockwise crossing
  //Increment the azimuth offset accordingly
  if (azDiff < -180) *azOffset += 360;
  if (azDiff > 180) *azOffset -= 360;

  //Save the current azimuth reading for the next iteration
  *azLast = az;

  //Compute the azimuth wind-up angle, i.e. the absolute number of degrees from the startup home position
  *azWindup = az + *azOffset;

  //Detect a windup condition where the antenna has rotated more than 450 degrees from home
  if (abs(*azWindup) > WINDUP_LIMIT) {
    *windup = true;    //Set the windup condition - it is reset later when the antenna nears home
  }

  //Perform the anti-windup procedure at the end of each pass
  //  if (elSet == 0)
  //    if (elLast > 0)
  //      if (mode == tracking) {
  //        *windup = true;
  //      }
  *elLast = elSet;
}

void getAzElDemo(float *azSet, float *elSet, float *azInc, float *elInc) {
  if (*azSet > 180.0) *azInc = -*azInc;
  if (*azSet < -180.0) *azInc = -*azInc;
  if (*elSet > 90.0) *elInc = -*elInc;
  if (*elSet < 0.0) *elInc = -*elInc;
  *azSet += *azInc;
  *elSet += *elInc;
  Serial1.print(*azSet, 0); Serial1.print(",");
  Serial1.println(*elSet, 0);
}

void getAzElError(float *azError, float *elError, bool *windup, float *azSet, float elSet, float az, float el) {
  //Compute the azimuth and elevation antenna pointing errors, i.e. angular offsets from set positions
  //Compute the azimuth antenna pointing error: Normally via the shortest path; opposite if windup detected.
  if (*windup) {                       //Check for a windup condition
    //To unwind the antenna set an azError in the appropriate direction to home
    *azError = constrain(azWindup, -180, 180); //Limit the maximum azimuth error to -180..180 degrees
    //Cancel the windup condition when the antenna is within 180 degrees of home
    //Set a zero home position by default, but return azumith control to the computer if still connected
    if (abs(*azError) < 90) {
      *windup = false;                 //Cancel windup and permit computer control
      *azSet = 0;                      //Set a default home position for last leg
    }
  }
  else {
    //Compute the normal azimuth antenna pointing error when there is no windup condition
    *azError = diffAngle(az, *azSet);
  }

  //Compute the elevation antenna pointing error
  *elError = diffAngle(el, elSet);
}

void processPosition() {
  //Perform the main operation of positioning the rotator under different modes
  //Read the accelerometer and magnetometer
  lsm303d.read();
  switch (mode) {
    case debugging:
      printDebug(); //Print the raw sensor data for debug purposes
      break;
    case calibrating:
      calibrate();  //Print calibration data
      break;
    default:
      getAzEl(&az, &el);                                                          //Get the antenna AZ and EL
      getWindup(&windup, &azWindup, &azOffset, &azLast, &elLast, az, elSet);      //Get the AZ windup angle and windup state
      if (mode == demonstrating) getAzElDemo(&azSet, &elSet, &azInc, &elInc);
      float azError, elError;
      getAzElError(&azError, &elError, &windup, &azSet, elSet, az, el);
      if (mode == monitoring) printMon(az, el, azSet, elSet, azWindup, azError, elError);
      //Drive the motors to reduce the azimuth and elevation error to zero
      drive(azError, azGain, &azSpeed, azBrkPin, azDirPin, azPwmPin);
      drive(elError, elGain, &elSpeed, elBrkPin, elDirPin, elPwmPin);
      break;
  }
}

void processUserCommands(String line) {
  //Process user commands
  //User command type 1: r, m, c, a, d or e <decl> followed by a carriage return
  //User command type 2: <az> <el> followed by a carriage return
  String param;                                           //Parameter value
  int firstSpace;                                         //Position of the first space in the command line
  int secondSpace;                                        //Position of the second space in the command line
  char command = line.charAt(0);
  switch (command) {                                      //Process type 1 user commands
    case 'r':                                             //Reset command
      Serial1.println("Reset in progress");
      reset(true);  //Reset the rotator and load configuration from flash
      Serial1.println("Reset");
      break;
    case 'b':                                             //Debug command
      Serial1.println("Debugging in progress: Press 'a' to abort");
      mode = debugging;
      t = 100;
      break;
    case 'm':                                             //Debug command
      Serial1.println("Monitoring in progress: Press 'a' to abort");
      mode = monitoring;
      t = 100;
      break;
    case 'c':                                             //Calibrate command
      Serial1.println("Calibration in progress: Press 'a' to abort or 's' to save");
      reset(false); //Reset the rotator, but don't load configuration from flash
      mode = calibrating;
      t = 50;
      break;
    case 'a':                                             //Abort command
      mode = tracking;
      t = 100;
      reset(true);
      Serial1.println("Function aborted");
      break;
    case 'e':
      param = line.substring(1);                          //Get the second parameter
      D = param.toFloat();
      break;
    case 's':                                             //Save command
      save(52, (byte*) &gs);
      reset(true);
      Serial1.println("Calibration saved");
      break;
    case 'd':
      Serial1.println("Demo in progress: Press 'a' to abort");
      t = 25;
      mode = demonstrating;
      break;
    default:                                              //Process type 2 user commands
      firstSpace = line.indexOf(' ');                     //Get the index of the first space
      param = line.substring(0, firstSpace);              //Get the first parameter
      azSet = param.toFloat();                            //Get the azSet value
      param = line.substring(firstSpace + 1);             //Get the second parameter
      elSet = param.toFloat();                            //Get the elSet value
  }
}

void processEasycommCommands(String line) {
  //Process Easycomm II rotator commands
  //Easycomm II position command: AZnn.n ELnn.n UP000 XXX DN000 XXX\n
  //Easycomm II query command: AZ EL \n
  String param;                                           //Parameter value
  int firstSpace;                                         //Position of the first space in the command line
  int secondSpace;                                        //Position of the second space in the command line
  if (line.startsWith("AZ EL")) {                         //Query command received
    printAzEl();                                          //Send the current Azimuth and Elevation
  }
  else {
    if (line.startsWith("AZ")) {                          //Position command received: Parse the line.
      firstSpace = line.indexOf(' ');                     //Get the position of the first space
      secondSpace = line.indexOf(' ', firstSpace + 1);    //Get the position of the second space
      param = line.substring(2, firstSpace);              //Get the first parameter
      azSet = param.toFloat();                            //Set the azSet value
      if (azSet > 180) azSet = azSet - 360;               //Convert 0..360 to -180..180 degrees format
      param = line.substring(firstSpace + 3, secondSpace);//Get the second parameter
      elSet = param.toFloat();                            //Set the elSet value
    }
  }
}

void processCommands(void) {
  //Process incoming data from the control computer
  //User commands are entered by the user and are terminated with a carriage return
  //Easycomm commands are generated by a tracking program and are terminated with a line feed
  while (Serial1.available() > 0) {
    char ch = Serial1.read();                                    //Read a single character from the serial buffer
    switch (ch) {
      case 13:                                                  //Carriage return received
        processUserCommands(line);                              //Process user commands
        line = "";                                              //Command processed: Clear the command line
        break;
      case 10:                                                  //Line feed received
        processEasycommCommands(line);                          //Process Easycomm commands
        line = "";                                              //Command processed: Clear the command line
        break;
      default:                                                  //Any other character received
        line += ch;                                             //Add this character to the command line
        break;
    }
  }
}

void setup() {
  //Set motor pins to outputs
  pinMode(azBrkPin, OUTPUT);
  pinMode(azDirPin, OUTPUT);
  pinMode(azPwmPin, OUTPUT);
  pinMode(elBrkPin, OUTPUT);
  pinMode(elDirPin, OUTPUT);
  pinMode(elPwmPin, OUTPUT);
  //Reset the rotator and load configuration from flash
  reset(true);
  //Initialize the serial port
  Serial1.begin(9600);
  //Initialize the sensor
  lsm303d.begin();
}

void loop() {
  processCommands();                                            //Process commands from the control computer
  processPosition();                                            //Process position feedback from the sensors
  delay(t);                                                     //Wait an appropriate time depending on the mode
}

